#!/usr/bin/env python
# -*- coding: utf-8 -*-

# # 2A.ml - Pipeline pour un réduction d'une forêt aléatoire - correction
# 
# Le modèle Lasso permet de sélectionner des variables, une forêt aléatoire produit une prédiction comme étant la moyenne d'arbres de régression. Cet aspect a été abordé dans le notebook [Reduction d'une forêt aléatoire](http://www.xavierdupre.fr/app/ensae_teaching_cs/helpsphinx/notebooks/td2a_tree_selection_correction.html). On cherche à automatiser le processus.

# In[1]:


from jyquickhelper import add_notebook_menu
add_notebook_menu()


# In[2]:


get_ipython().run_line_magic('matplotlib', 'inline')


# ## Datasets
# 
# Comme il faut toujours des données, on prend ce jeu [Diabetes](https://scikit-learn.org/stable/modules/generated/sklearn.datasets.load_diabetes.html).

# In[3]:


from sklearn.datasets import load_diabetes
data = load_diabetes()
X, y = data.data, data.target


# In[4]:


from sklearn.model_selection import train_test_split
X_train, X_test, y_train, y_test = train_test_split(X, y)


# ## Forêt aléatoire suivi de Lasso
# 
# La méthode consiste à apprendre une forêt aléatoire puis à effectuer d'une régression sur chacun des estimateurs.

# In[5]:


import numpy
from sklearn.ensemble import RandomForestRegressor
from sklearn.linear_model import Lasso

# Apprentissage d'une forêt aléatoire
clr = RandomForestRegressor()
clr.fit(X_train, y_train)

# Récupération de la prédiction de chaque arbre
X_train_2 = numpy.zeros((X_train.shape[0], len(clr.estimators_)))
estimators = numpy.array(clr.estimators_).ravel()
for i, est in enumerate(estimators):
    pred = est.predict(X_train)
    X_train_2[:, i] = pred

# Apprentissage d'une régression Lasso
lrs = Lasso(max_iter=10000)
lrs.fit(X_train_2, y_train)
lrs.coef_


# Nous avons réussi à reproduire le processus dans son ensemble. Pas toujours simple de se souvenir de toutes les étapes, c'est pourquoi il est plus simple de compiler l'ensemble dans un [pipeline](https://scikit-learn.org/stable/modules/generated/sklearn.pipeline.Pipeline.html).

# ## Premier pipeline
# 
# L'idée est d'avoir quelque chose qui ressemble à ce qui suit.

# In[6]:


from sklearn.pipeline import Pipeline

try:
    pipe = Pipeline(steps=[
        ('rf', RandomForestRegressor()),
        ("une fonction qui n'existe pas encore", fct),
        ("lasso", Lasso()),
    ])
except Exception as e:
    print(e)


# Dans un pipeline, on ne peut y mettre que des modèles prédictifs, classifieur, régresseur ou des transformeur (normalisseur). La fonction qui extrait les prédictions des arbres doit être emballés dans un *transformer*. C'est le rôle d'un [FunctionTransformer](https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.FunctionTransformer.html).

# In[7]:


from sklearn.preprocessing import FunctionTransformer

def random_forest_tree_prediction(rf, X):
    preds = numpy.zeros((X.shape[0], len(rf.estimators_)))
    estimators = numpy.array(rf.estimators_).ravel()
    for i, est in enumerate(estimators):
        pred = est.predict(X)
        preds[:, i] = pred
    return preds
    

random_forest_tree_prediction(clr, X)


# In[8]:


fct = FunctionTransformer(lambda X, rf=clr: random_forest_tree_prediction(rf, X) )

fct.transform(X_train)


# Tout se passe bien. Il suffit de l'insérer dans le pipeline.

# In[9]:


try:
    pipe = Pipeline(steps=[
        ('rf', RandomForestRegressor()),
        ("tree_pred", fct),
        ("lasso", Lasso()),
    ])
except Exception as e:
    print(e)


# Ca ne marche toujours pas parce qu'un pipeline à ce que toutes les étapes excepté la dernière doivent être un *transformeur* et implémenter la méthode ``transform`` et ce n'est pas le cas. Et cela pose également un autre problème, la fonction ne fonctionne que si elle reçoit la forêt aléatoire en argument et nous avons passé celle déjà apprise mais ce n'aurait pas été celle apprise dans le pipeline.

# In[10]:


hasattr(clr, 'transform')


# In[11]:


from jyquickhelper import RenderJsDot


# In[12]:


RenderJsDot("""digraph {
  A [label="RandomForestRegressor pipline"];
  A2 [label="RandomForestRegressor - pretrained"];
  B [label="FunctionTransformer"]; C [label="Lasso"];
  A -> B [label="X"]; B -> C [label="X2"]; A2 -> B [label="rf"]; }""")


# Comme ça ne marche pas, on passe à une seconde idée.

# ## Second pipeline
# 
# On déguise la forêt aléatoire en un transformeur.

# In[13]:


class RandomForestRegressorAsTransformer:
    
    def __init__(self, **kwargs):
        self.rf = RandomForestRegressor(**kwargs)
        
    def fit(self, X, y):
        self.rf.fit(X, y)
        return self
        
    def transform(self, X):
        preds = numpy.zeros((X.shape[0], len(self.rf.estimators_)))
        estimators = numpy.array(self.rf.estimators_).ravel()
        for i, est in enumerate(estimators):
            pred = est.predict(X)
            preds[:, i] = pred
        return preds


trrf = RandomForestRegressorAsTransformer()
trrf.fit(X_train, y_train)
trrf.transform(X_train)


# Tout va bien. On refait le pipeline.

# In[14]:


pipe = Pipeline(steps=[('trrf', RandomForestRegressorAsTransformer()),
                       ("lasso", Lasso())])

pipe.fit(X_train, y_train)


# On récupère les coefficients.

# In[15]:


pipe.steps[1][1].coef_


# ## A quoi ça sert : GridSearchCV
# 
# Comme l'ensemble des traitements sont maintenant dans un seul pipeline que *scikit-learn* considère comme un modèle comme les autres, on peut rechercher les meilleurs hyper-paramètres du modèle, comme le nombre d'arbres initial, le paramètre *alpha*, la profondeur des arbres...

# In[16]:


from sklearn.model_selection import GridSearchCV

param_grid = {'trrf__n_estimators': [30, 50, 80, 100],
              'lasso__alpha': [0.5, 1.0, 1.5]}

try:
    grid = GridSearchCV(pipe, cv=5, verbose=1, param_grid=param_grid)
    grid.fit(X_train, y_train)
except Exception as e:
    print(e)


# La classe ``RandomForestRegressorAsTransformer`` a besoin de la méthode *set_params*... Aucun problème.

# In[17]:


class RandomForestRegressorAsTransformer:
    
    def __init__(self, **kwargs):
        self.rf = RandomForestRegressor(**kwargs)
        
    def fit(self, X, y):
        self.rf.fit(X, y)
        return self
        
    def transform(self, X):
        preds = numpy.zeros((X.shape[0], len(self.rf.estimators_)))
        estimators = numpy.array(self.rf.estimators_).ravel()
        for i, est in enumerate(estimators):
            pred = est.predict(X)
            preds[:, i] = pred
        return preds
    
    def set_params(self, **params):
        self.rf.set_params(**params)


# In[18]:


import warnings
from sklearn.exceptions import ConvergenceWarning

pipe = Pipeline(steps=[('trrf', RandomForestRegressorAsTransformer()),
                       ("lasso", Lasso())])
    
param_grid = {'trrf__n_estimators': [50, 100],
              'lasso__alpha': [0.5, 1.0, 1.5]}

grid = GridSearchCV(pipe, cv=5, verbose=2, param_grid=param_grid)

with warnings.catch_warnings(record=False) as w:
    # On ignore les convergence warning car il y en beaucoup.
    warnings.simplefilter("ignore", ConvergenceWarning)
    grid.fit(X_train, y_train)


# In[19]:


grid.best_params_


# In[20]:


grid.best_estimator_.steps[1][1].coef_


# In[21]:


grid.best_score_


# On essaye sur la base de test.

# In[22]:


grid.score(X_test, y_test)


# Et il y a combien de coefficients non nuls extactement...

# In[23]:


coef = grid.best_estimator_.steps[1][1].coef_
coef.shape, sum(coef != 0)


# Et on ne peut pas vérifier pour les autres car la recherche ne conserve pas les autres modèles. Et il reste un problème qui est plus ou moins important selon l'usage qu'on fait du modèle : il y a 23 coefficients nuls ce qui fait le modèle calcule la prédiction de 23 arbres de régressions qui ne sont pas utilisées puisque la régression leur affecte un coefficient nul. On pourrait donc facilement diviser le temps de prédiction par deux. Ce n'est pas si difficile à faire mais ce sera pour une autre histoire.

# In[24]: